## Example code for BENG 260 Computational Lab and Homework 4, Fall 2019
## Questions to mwagner@ucsd.edu, jmford@ucsd.edu or gert@ucsd.edu

import scipy as sp
import pylab as plt
from scipy.integrate import odeint
from ndtools.spikes import *
import numpy as np


## Hodgkin-Huxley Model

# Constants
C_m  =   1.0 # membrane capacitance, in uF/cm^2
g_Na = 120.0 # maximum conducances, in mS/cm^2
g_K  =  36.0
g_L  =   0.3
E_Na =  45.0 # Nernst reversal potentials, in mV
E_K  = -82.0
E_L  = -59.387
E_Cl = -80.0
E_excite = -38.0

# Channel gating kinetics
# Functions of membrane voltage
def alpha_m(V): return 0.1 * (V + 45.0) / (1.0 - sp.exp(-(V + 45.0) / 10.0))
def beta_m(V):  return 4.0 * sp.exp(-(V + 70.0) / 18.0)
def alpha_h(V): return 0.07 * sp.exp(-(V + 70.0) / 20.0)
def beta_h(V):  return 1.0 / (1.0 + sp.exp(-(V + 40.0) / 10.0))
def alpha_n(V): return 0.01 * (V + 60.0)/(1.0 - sp.exp(-(V + 60.0) / 10.0))
def beta_n(V):  return 0.125 * sp.exp(-(V + 70.0) / 80.0)

# Alpha / Beta constants for inhibitory / excitatory synapses
alpha_r_inhibit = 5.0   # 1/(mM*ms)
beta_r_inhibit  = 0.18  # 1/ms
alpha_r_excite = 2.4   # 1/(mM*ms)
beta_r_excite  = 0.56  # 1/ms

# [T] equation for synapses
T_max_inhibit = 1.5   # mM (inhibitory)
T_max_excite = 1.0   # mM (excitatory)
K_p = 5.0       # mV
V_p = 7.0       # mV
def T_inhibit(V_pre): return T_max_inhibit / (1.0 + sp.exp(-(V_pre - V_p) / K_p))
def T_excite(V_pre): return T_max_excite / (1.0 + sp.exp(-(V_pre - V_p) / K_p))

# Differential gating equations
def dm(V, m):   return alpha_m(V) * (1.0 - m) - beta_m(V) * m
def dh(V, h):   return alpha_h(V) * (1.0 - h) - beta_h(V) * h
def dn(V, n):   return alpha_n(V) * (1.0-n) - beta_n(V) * n
def dr_inhibit(V, r):
    return alpha_r_inhibit * T_inhibit(V) * (1.0 - r) - beta_r_inhibit * r # V=V_pre, r is for pre
def dr_excite(V, r): return alpha_r_excite * T_excite(V) * (1.0 - r) - beta_r_excite * r # V=V_pre, r is for pre

# Membrane currents (in uA/cm^2)
def I_Na(V, m, h): return g_Na * m**3 * h * (V - E_Na)
def I_K(V, n):   return g_K * n**4 * (V - E_K)
def I_L(V):      return g_L * (V - E_L)
def I_syn_inhibit(V, g_GABA, r): return sp.dot(r,g_GABA) * (V - E_Cl) # V=V_post, r is for pre
def I_syn_excite(V, g_Glu, r):  return sp.dot(r,g_Glu ) * (V - E_excite) # V=V_post, r is for pre

# Membrane voltage differential
def dV(V, m, h, n, r_inhibit, r_excite, I_ext, g_GABA, g_Glu):
    # Voltage function
    return (-I_Na(V, m, h) - I_K(V, n) - I_L(V) - I_syn_inhibit(V, g_GABA, r_inhibit) - I_syn_excite(V, g_Glu, r_excite) + I_ext) / C_m


def d_single(hh_vars, t, I_ext, g_GABA, g_Glu):
    '''
    Differential equations for a single neuron. This function works for a single cell when using odeint, but for more than
    one cell you need to use the 'd' function below.
    :param hh_vars:     V, m, h, n, r_inhibit, and r_excite.
    :param t:           time array
    :param I_ext:       applied external current; accepts the time and returns a row vector of current applied to each cell.
    :param g_GABA:      matrix of inhibitory synapses where each row is the presynaptic cell and the column is the postsynaptic cell.
    :param g_Glu:       matrix of excitatory synapses where each row is the presynaptic cell and the column is the postsynaptic cell.
    :return:            differential equation rhs
    '''
    V, m, h, n, r_inhibit, r_excite = hh_vars
    return np.array([
        dV(V, m, h, n, r_inhibit, r_excite, I_ext(t), g_GABA, g_Glu),
        dm(V, m),
        dh(V, h),
        dn(V, n),
        dr_inhibit(V, r_inhibit),
        dr_excite(V, r_excite)
        ])

def d(hh_vars, t, I_exts, g_GABA, g_Glu):
    '''
    Differential equation rhs for many neurons. This uses the single version, which is vectorizable, however since odeint
    needs a 1D array this function reshapes the vector to work properly.
    :param hh_vars:     V, m, h, n, r_inhibit, and r_excite.
    :param t:     time array
    :param I_ext:       applied external current; accepts the time and returns a row vector of current applied to each cell.
    :param g_GABA:      matrix of inhibitory synapses where each row is the presynaptic cell and the column is the postsynaptic cell.
    :param g_Glu:       Same for g_Glu except that it is excitatory.
    :return:            differential equation rhs
    '''
    return np.reshape(d_single(sp.reshape(hh_vars, (6, len(hh_vars)//6)), t, I_exts, g_GABA, g_Glu), len(hh_vars))


t = np.arange(0, 500, 0.1)
def network(I_exts, g_GABA, g_Glu):
    '''
    # Shortcut for simulating a whole network

    :param I_exts:      applied external current; accepts the time and returns a row vector of current applied to each cell.
    :param g_GABA:      matrix of inhibitory synapses where each row is the presynaptic cell and the column is the postsynaptic cell.
    :param g_Glu:       matrix of excitatory synapses where each row is the presynaptic cell and the column is the postsynaptic cell.
    :return:        3D matrix where the first dimension is the variable (V,m,h,n,r_inhibit, and r_excite),
    second dimension is the cell, and third dimension is over time
    '''
    ncells = np.size(g_GABA, 0)
    return np.reshape(
        odeint(d, sp.repeat([-70, 0.05, 0.6, 0.3, 0, 0], ncells), t, (I_exts, g_GABA, g_Glu)).T,
        (6, ncells, len(t)))

# Convenience variable for no connections for a 2 cell network
nocon2 = sp.zeros((2,2))
